C
C cvdbg.f:  Test, debug and time convolution routines
C	The debug routines exercise the direct and FFT convolution
C	routines for smooth if and random charges, and they report
C	the integral difference between the results (which is only
C	numerical noise in the last digits as of Jan. 3rd).
C	CVTIMING can be used to measure machine performance.
C	CVTRANS has been used to find the optimum order to transpose
C	the arrays in CV3D.
C	CVFFTTIM measures FFT performance as function of input array
C	length 
C Authors: P. Lockey and K. Ramstoeck, Oct. 17th 1995
C
      Program CVDBG
      implicit none
      integer choice
c
c      call CVTIMING
c      stop
c
 10     continue
        print 1000, 'CV-debugger:'
        print 1000, '1 - debug cv1d'
        print 1000, '2 - debug cv2d'
        print 1000, '3 - debug cv3d'
        print 1000, '4 - Time CV2D'
        print 1000, '5 - Time CV3D'
        print 1000, '6 - Time transpose'
        print 1000, '7 - Time FFT'
        print 1000, 'Enter 1 ... 7 (0 to exit):'
        read(*,*) choice
        if (choice.eq.1) call CVDBG1D
        if (choice.eq.2) call CVDBG2D
        if (choice.eq.3) call CV3DDBG
        if (choice.eq.4) call CVTIMING
        if (choice.eq.5) call CVTIMING3D
        if (choice.eq.6) call CVTRANS
        if (choice.eq.7) call CVFFTTIM
        if (choice.eq.0) go to 100
        go to 10

100	print 1000, 'CVDBG beendet.'
1000	format (A)
        end
c
c
c
      Subroutine CVDBG2D
      implicit none
      include 'common.inc'
      real*8 ww(NM,MM),wwft(4*NM*MM),wwd(NM*MM),fi(0:NM+1,0:MM+1)
      real*8 ch1(NM, MM), ch2(NM, MM), err, sc(1)
      real*8 rand
      external rand
      integer i,j,sz,h,k,sz2,N,M,ct
c
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ ww,wwft,wwd,fi,ch1,ch2
c
      call RandSeed(4711.123D0)
      h=0
      k=0
      print *, 'CVDBG2D - N, M'
      read(*,*) N, M
c
 10   continue
c
c Initialize ch(*) and ww:
      do 100 j=1, MM
         do 100 i=1, NM
            ch1(i,j)=4E3
            ch2(i,j)=4E3
            ww(i,j)=4E4
            wwft((j-1)*NM+i)=4E5
            wwd((j-1)*NM+i)=4E6
 100  continue

      do 105 j=0, MM+1
         do 105 i=0, NM+1
            fi(i,j)=3.333
 105     continue

c
      print *, 'Cutoff:'
      read(*,*) ct
      do 110 j=1, M
         do 110 i=1, N
            ch1(i,j)= rand()
            ch2(i,j)= ch1(i,j)
            ww(i,j)= 1/(i+j-1.0D0)
            wwft((j-1)*NM+i)=0
            if(i.gt.n-ct) ww(i,j)=0
            if(j.gt.M-ct) ww(i,j)=0
 110  continue
c
      sz=4*NM*MM
      sz2=NM*MM
c      print *, 'h,k:'
c      read(*,*) h,k
c      ww(1,1)=1
c      ww(1,2)=1
c      ww(2,1)=1
c      ww(3,1)=1
c      ww(1,3)=1
c      ch1(h,k)=1
c      ch2(h,k)=1
      call PrtArray(ch1, N, M, NM, 'ch:')
      call PrtArray(ww, N, M, NM, 'ww:')
c
      call CV2DInit(ww, N, M, NM, wwft, sz)
      if(sz.lt.0) then
         print *, 'CV2DInit failed with sz = ', sz
         stop
      else
         print *, 'CV2DInit succeeded with sz=', sz
      endif
c      call PrtArray(wwft, M*2, N*2, M*2, 'wwft:')
c
      call CV2DDInit(ww, N, M, NM, wwd, sz2)
      if(sz2.lt.0) then
         print *, 'CV2DDInit failed with sz = ', sz2
         stop
      else
         print *, 'CV2DDInit succeeded with sz2 =', sz2
      endif
c      call PrtArray(wwd, N, M, N, 'wwd:')
c
      call CV2D(ch1,N,M,NM,wwft,fi,2,2,NM+2,sc,0 )
c
      call PrtArray(fi, N+2, M+2, NM+2, 'Final Result-FFT:')
c
      call CV2DD(ch2, N, M, NM, wwd, ch2)
c      call CV2DDInit(ww, N, M, NM, wwd, sz2)
c
      call PrtArray(ch2, N, M, NM, 'Final Result-direct:')
c
      err=0
      do 1000 j=1, M
         do 1000 i=1, N
            err = err+(fi(i,j)-ch2(i,j))**2
            ch2(i,j)=ch2(i,j) - fi(i,j)
 1000 continue
      print *, 'Avg. Error: ', sqrt(err)/(N*M)
      call PrtArray(ch2, N, M, NM, 'Error distribution:')
c
c      go to 10
c
      end
c
c
c
      Subroutine CVDBG1D
      implicit none
      include 'common.inc'
      real*8 ww(NM,MM), wwft(4*NM*MM), wwd(NM*MM), fi(0:NM+1,0:MM+1)
      real*8 ch1(NM, MM), ch2(NM, MM), err, sc(1)
      integer i,j,sz,h,k,sz2,N,M
c
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ ww,wwft,wwd,fi,ch1,ch2
c
      h=0
      k=0
      print *, 'CVDBG1D - N, M'
      read(*,*) N, M
c
 10   continue
c
c Initialize ch(*) and ww:
      do 100 j=1, MM
         do 100 i=1, NM
            ch1(i,j)=1E3
            ch2(i,j)=1E3
            ww(i,j)=1E4
            wwft((j-1)*NM+i)=1E5
            wwd((j-1)*NM+i)=1E6
 100  continue
      do 105 j=0, MM+1
         do 105 i=0, NM+1
            fi(i,j)=3.333
 105     continue

c
      do 110 j=1, M
         do 110 i=1, N
            ch1(i,j)= i+j
            ch2(i,j)= i+j
            ww(i,j)=0
            if(i.le.j) ww(i,j)=1
            wwft((j-1)*NM+i)=0
 110  continue
c
      sz=4*NM*MM
      sz2=NM*MM
c      print *, 'h,k:'
c      read(*,*) h,k
      ww(1,1)=1
c      ww(1,2)=1
c      ww(2,1)=1
c      ww(2,2)=1
c      ch1(h,k)=1
c      ch2(h,k)=1
      call PrtArray(ch1, N, M, NM, 'ch:')
      call PrtArray(ww, N, M, NM, 'ww:')
c
      call CV1DInit(ww, N, M, NM, wwft, sz)
      if(sz.lt.0) then
         print *, 'CV1DInit failed with sz = ', sz
         stop
      else
         print *, 'CV1DInit succeeded with sz=', sz
      endif
      call PrtArray(wwft, N*4, M, N*4, 'wwft:')
c
      call CV1DDInit(ww, N, M, NM, wwd, sz2)
      if(sz2.lt.0) then
         print *, 'CV1DDInit failed with sz2 = ', sz2
         stop
      else
         print *, 'CV1DDInit succeeded with sz2 =', sz2
      endif
c      call PrtArray(wwd, N, M, N, 'wwd:')
c
      call CV1D(ch1,N,M,NM,wwft,M,fi,2,2,NM+2,sc,0 )
c
      call PrtArray(fi, N+2, M+2, NM+2, 'Final Result-FFT:')
c
      call CV1DD(ch2, N, M, NM, wwd, M)
c
      call PrtArray(ch2, N, M, NM, 'Final Result-direct:')
c
      err=0
      do 1000 j=1, M
         do 1000 i=1, N
            err = err+(fi(i,j)-ch2(i,j))**2
 1000 continue
      print *, 'Avg. Error: ', sqrt(err)/(N*M)
c
c      go to 10
c
      end
c
      Subroutine PrtArray(d, N, M, NM, lab)
      implicit none
      integer N, M, NM
      real*8 d(1:NM, 1:M)
      character*(*) lab
      integer i,j
c
        print *, lab
	if(N.gt.8 .or. M.gt.8) return
        do 120 j=1, M
           write(*, 1000) (d(i,j),i=1, N)
 120    continue
c
 1000 format(256G9.2)
        return
      end
c
c
c
      Subroutine PrtArray3D(d, H,K,L, HM,KM, lab)
      implicit none
      integer H,K,L, HM,KM
      real*8 d(HM,KM,L)
      character*(*) lab
      integer p,q,r
c
        print *, lab
	if(H.gt.8 .or. K.gt.8 .or. L.gt.8) return
        do 120 r=1, L
           print *, 'r = ', r
           do 120 q=1, K
              write(*, 1000) (d(p,q,r),p=1, H)
 120    continue
c
 1000 format(256G9.2)
        return
      end
c
c
c
      Subroutine CVTiming
c Exercise the CV-routines to allow for timing
      implicit none
      include 'common.inc'
      real*8 ww(NM,MM),wwft((NM+1)*(MM+1)),fi(0:NM+1,0:MM+1),fid(NM,MM)
      real*8 ch(NM, MM), sc(1), tim, ch2(NM, MM), wwd(NM*MM)
      real*8 rand, err, timer, delay, ddelay, erc, errm
      character*10 fnam, fout
      external rand, timer
      integer i,j,sz,k,sz2,N,M, rep,drep
c
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ ww,wwft,wwd,fi,ch,ch2
c
      call RandSeed(4711.123D0)
      call pStart()

      fnam='CVINP.txt'
      fout='CVOUT.txt'
      print '(4A)','CVTiming - reading from ',fnam,' writing to ',fout
      open(unit=10, file=fnam, status='old')
      open(unit=11, file=fout, status='unknown')
      write(11,'(A,A)') '#  Z=N*M    N    M',
     &	'   rep     dfft/rep       ddir         err            errm'

 500  continue
      read(10,*, end=2000) N, M, rep
      if(N.le.0) go to 2000

      print *, ' '
      print 4711, 'Working on ',N, ' x ', M, ' for ',rep,' times.'
 4711 format (A,I4, A, I4, A, I6, A)

        do 10 j=1, M
           do 10 i=1, N
              ch(i,j) = rand()
              ww(i,j) = 1.0/(i+j-1)
 10     continue
c
c
        sz=(NM+1)*(MM+1)	! Minimum required size for NM*MM preblem

        call CV2DInit(ww, N, M, NM, wwft, sz)
        if(sz.lt.0) then
          print *, 'CV2DInit failed with sz = ', sz
          stop
         else
c           print *, 'CV2DInit succeeded with sz=', sz
        endif
c
       tim=timer()

      do 1000 k=1, rep
        call CV2D(ch,N,M,NM,wwft,fi,2,2,NM+2,sc,0 )
 1000   continue
 	delay = timer()-tim
        print 4712, 'This took: ',delay , ' [s] = ',
     &		delay/rep, ' [s] per conv.'
 4712	format (A, G9.3, A, G9.3, A)

c Verify result:

        print '(A)', 'Verifying results:'
        sz2=NM*MM
        call CV2DDInit(ww, N, M, NM, wwd, sz2)
        if(sz2.lt.0) then
          print *, 'CV2DDInit failed with sz = ', sz2
          stop
         else
c           print *, 'CV2DDInit succeeded with sz=', sz2
        endif

        tim = timer()
        drep=0
 150	continue        
 	drep=drep+1
        call CV2DD(ch, N, M, NM, wwd, fid)
        ddelay = timer()-tim
        if( ddelay.lt.5.0D0 ) go  to 150	! Loop at least 1 sec.
        ddelay=ddelay/drep
        
        print 4713, 'This took: ',ddelay*drep, ' [s] (',drep,
     &		' cycles) ',ddelay, ' [s] per conv.'
 4713	format (A, G9.3, A,I4,A, G9.3,A)
        print 4714, 'Speedup by fft: ', (ddelay*rep)/delay
 4714 	format (A, G9.3)
c
        err=0
        errm=0
        do 200 j=1, M
           do 200 i=1, N
              erc=(fid(i,j)-fi(i,j))
              erc=erc*erc
              err=err+erc
              errm=MAX(erc,errm)
 200    continue
 	err=sqrt(err)/(N*M)
 	errm=sqrt(errm)
        print 5100, 'Avg. deviation direct/fft convolution: ', err
        print 5100, 'Max. deviation direct/fft convolution: ', errm
 5100	format (A, G12.6)      
        write(11,5000) N*M, N, M, rep, delay/rep, ddelay, err, errm
 5000   format (I8, 2I5, I6, 4(' ',G12.6,' '))

        go to 500

 2000   continue

        close(10)
        close(11)
 
        return
      end
c
c
c
      Subroutine CVTiming3D
c Exercise the CV-routines to allow for timing, 3D case
      implicit none
      integer HM,KM,LM
      parameter (HM=32, KM=32, LM=32)
      real*8 lam(HM,KM,LM), ww(0:HM,0:KM,0:LM)
      real*8 wwd(HM,KM,LM),fif(0:HM+1,0:KM+1,0:LM+1)
      real*8 fid(HM,KM,LM)
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ lam,ww,wwd,fif,fid
      real*8 err, delay, ddelay, erc, errm, tim, sc(1)
      character*12 fnam, fout
      real*8 rand, timer
      external rand, timer
      integer sz,sz2, rep,drep, p,q,r, h,k,l
c
      call RandSeed(4711.123D0)
      call pStart()

      fnam='CVINP3D.txt'
      fout='CVOUT3D.txt'
      print '(4A)','CVTiming - reading from ',fnam,' writing to ',fout
      open(unit=10, file=fnam, status='old')
      open(unit=11, file=fout, status='unknown')
      write(11,'(A,A)') '# Z=H*K*L   H    K    L',
     &	'  rep     dfft/rep       ddir         err            errm'

 500  continue
      read(10,*, end=2000) H,K,L, rep
      if(H.le.0) go to 2000

      print *, ' '
      print 4711,'Working on ',H,' x ',K,' x ',L,' for ',rep,' times.'
 4711 format (A,I4, A, I4, A, I4, A, I6, A)

        do 10 p=1, L
          do 10 q=1, K
            do 10 r=1, H
              lam(r,q,p) = rand()
               ww(r,q,p) = 1.0/(r+p+q-1)
 10     continue
c
c
        sz2=HM*KM*LM
        call CV3DDInit(ww, H,K,L, HM,KM, wwd, sz2)
        if(sz2.lt.0) then
          print *, 'CV3DDInit failed with sz = ', sz2
          stop
         else
c           print *, 'CV3DDInit succeeded with sz=', sz2
        endif
c
c Input- and Output array for CV3DInit can be the same!
c it is legal and sane to make sz bigger than needed- later versions
c of CVxD might use that to empoly efficient FFT-sizes
        sz=(HM+1)*(KM+1)*(LM+1)
        call CV3DInit(ww, H,K,L, HM,KM, ww, sz)
        if(sz.lt.0) then
          print *, 'CV3DInit failed with sz = ', sz
          stop
         else
c           print *, 'CV3DInit succeeded with sz=', sz
        endif
c
       tim=timer()

      do 1000 r=1, rep
        call CV3D(lam,H,K,L,HM,KM,ww,fif,2,2,2,HM+2,KM+2,sc,0 )
 1000 continue
 	delay = timer()-tim
        print 4712, 'This took: ',delay , ' [s] = ',
     &		delay/rep, ' [s] per conv.'
 4712	format (A, G9.3, A, G9.3, A)

c Verify result:

        print '(A)', 'Verifying results:'

        tim = timer()
        drep=0
 150	continue        
 	drep=drep+1
        call CV3DD(lam, H,K,L, HM,KM, wwd, fid)
        ddelay = timer()-tim
        if( ddelay.lt.5.0D0 ) go  to 150	! Loop at least 5 sec.
        ddelay=ddelay/drep
        
        print 4713, 'This took: ',ddelay*drep, ' [s] (',drep,
     &		' cycles) ',ddelay, ' [s] per conv.'
 4713	format (A, G9.3, A,I4,A, G9.3,A)
        print 4714, 'Speedup by fft: ', (ddelay*rep)/delay
 4714 	format (A, G9.3)
c
        err=0
        errm=0
        do 200 p=1, L
          do 200 q=1, K
            do 200 r=1, H
              erc=(fid(r,q,p)-fif(r,q,p))
	      erc=erc*erc
              err=err+erc
              errm=MAX(errm,erc)
 200    continue
 	err=sqrt(err)/(H*K*L)
 	errm=sqrt(errm)
        print 5100, 'Avg. deviation direct/fft convolution: ', err
        print 5100, 'Max. deviation direct/fft convolution: ', errm
 5100	format (A, G12.6)      
        write(11,5000) H*K*L, H,K,L, rep, delay/rep, ddelay, err, errm
 5000   format (I8, 3I5, I6, 4(' ',G12.6,' '))

        go to 500

 2000   continue

        close(10)
        close(11)
 
        return
      end
c
c
c
      Subroutine CVTrans
c Measure how fast 3D-transposing a matrix is
      implicit none
      include 'common.inc'
      real*8 in(MM*MM*MM), out(MM*MM*MM), tmp(MM*MM*MM)
      real*8 rand, err, tim, timer, delay, diff
      external rand, timer, diff
      integer i, HM,KM,LM, l,rep
c
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ in, out
c
       	call RandSeed(4711.123D0)
      	call pStart()

	HM=MM
	KM=MM
	LM=MM
	
	rep=100

      	print *, 'HM, KM, LM:', HM, KM, LM, ' Repeat:', rep
      	print *, 'Initializing source array ...'

      	do 100 i=1, HM*KM*LM
          in(i)=rand()
100     continue

	print *, 'Transpose of type d(HM,KM,LM) -> d(KM,HM,LM) (213)'
	tim=timer()
	do 200 l=1, rep
	  call tr213(In, Out, HM,KM,LM)
200	continue
	delay=(timer()-tim)/rep
	print *, 's/transpose: ', delay

	print *, 'Transpose of type d(HM,KM,LM) -> d(KM,LM,HM) (231)'
	tim=timer()
	do 300 l=1, rep
	  call tr231(In, Out, HM,KM,LM)
300	continue
	delay=(timer()-tim)/rep
	print *, 's/transpose: ', delay

	print *, 'Transpose of type d(HM,KM,LM) -> d(LM,HM,KM) (312)'
	tim=timer()
	do 400 l=1, rep
	  call tr312(In, Out, HM,KM,LM)
400	continue
	delay=(timer()-tim)/rep
	print *, 's/transpose: ', delay

	print *, 'Transpose of type d(HM,KM,LM) -> d(LM,KM,HM) (321)'
	tim=timer()
	do 500 l=1, rep
	  call tr321(In, Out, HM,KM,LM)
500	continue
	delay=(timer()-tim)/rep
	print *, 's/transpose: ', delay

	print *, 'Verification:'
	call tr213(In, Out, HM,KM,LM)
	call tr213(Out, tmp, KM,HM,LM)
	err = diff(tmp, In, HM*KM*LM)
	print *, 'tr213: ', err

	call tr231(In, Out, HM,KM,LM)
	call tr231(Out, tmp, KM,LM,HM)
	call tr231(tmp, Out, LM,HM,KM)
	err = diff(Out, In, HM*KM*LM)
	print *, 'tr231: ', err
	
	call tr312(In, Out, HM,KM,LM)
	call tr312(Out, tmp, LM,HM,KM)
	call tr312(tmp, Out, KM,LM,HM)
	err = diff(Out, In, HM*KM*LM)
	print *, 'tr312: ', err

	call tr321(In, Out, HM,KM,LM)
	call tr321(Out, tmp, LM,KM,HM)
	err = diff(tmp, In, HM*KM*LM)
	print *, 'tr321: ', err
	
        return
      end
c
c
c
      Subroutine CV3DDBG
c Debug the 3D-convolution routines
      implicit none
      integer HM,KM,LM
      parameter (HM=5, KM=5, LM=5)
      real*8 lam(HM,KM,LM), ww(0:HM-1,0:KM-1,0:LM-1), fid(HM,KM,LM)
      real*8 wwd(HM,KM,LM),fif(HM,KM,LM),wwf(HM*2,KM*2,LM*2), tmp
      real*8 err,errc, ddelay, fdelay,tim
      integer p,q,r,h, k,l, sz
      real*8 rand, timer
      external rand,timer
c
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ lam,ww,fid,wwd,fif,wwf
c
c
c Initialization
       	call RandSeed(4711.123D0)
c     Get Sizes:
      print *, 'H, K, L (<',HM,KM,LM,')'
      read(*,*) H,K,L
c
c     fill ww with values:
        do 10 r=0, L-1
           do 10 q=0, K-1
              do 10 p=0, H-1
                 ww(p,q,r) = r + q*q + 1/(p+1.0D0)
 10     continue
c 	ww(0,0,0)=1.234
c
c     blank out lam:
        do 20 r=1, L
           do 20 q=1, K
              do 20 p=1, H
                 lam(p,q,r) = 0.0D0
                 fid(p,q,r) = 0.0D0
                 fif(p,q,r) = 4712.0D0
 20     continue
c
c     enter lam values:
 30     continue
           print *, 'p,q,r, lam(p,q,r) or -1 (end) or -2 (random):'
           read(*,*,ERR=45,END=45) p,q,r,tmp
           if(p.eq.-2) go to 35
           if(p.lt.0) go to 45
           lam(p,q,r)=tmp
           fid(p,q,r)=tmp
        goto 30	
 35	continue
        do 40 r=1, L
           do 40 q=1, K
              do 40 p=1, H
                 lam(p,q,r) = rand()
                 fid(p,q,r) = lam(p,q,r)
 40     continue
 45     continue
c
c Prepare for convolution
	sz=H*K*L
        call CV3DDInit(ww, H,K,L, HM, KM, wwd, sz)
      	if(sz.lt.0) then
          print *, 'CV3DDInit failed.'
          return
      	endif
c
c Convolute direct:
	tim=timer()
        call CV3DD(fid, H,K,L, HM,KM, wwd, fid)
	ddelay=timer()-tim
	print *, 'sec / Direct convolution:', ddelay
c        call PrtArray3D(wwd, H,K,L, H,K, 'wwd:')
        call PrtArray3D(fid, H,K,L, HM,KM, 'fid:')
c
c     Blank out wwf:
        do 50 r=1, LM*2
           do 50 q=1, KM*2
              do 50 p=1, HM*2
                 wwf(p,q,r) =  0.0D0
 50     continue
c
c Initialize 3D-FFT convolution
      	sz=8*H*K*L
      	call CV3DInit(ww, H,K,L, HM,KM, wwf, sz)
c      	call PrtArray3D(wwf, H+1,K+1,L+1, H+1,K+1, 'wwf:')
      	if(sz.lt.0) then
          print *, 'CV3DInit failed.'
          return
      	endif
      	tim=timer()
      	call CV3D(lam, H,K,L, HM,KM, wwf, fif, 1,1,1, HM,KM, tmp,0)
      	fdelay=timer()-tim
	print *, 'sec / fft convolution:', fdelay
      call PrtArray3D(fif, H,K,L, HM,KM, 'fif:')
c
c     Compute Error: difference between fif and fid
	err=0.0D0
        do 60 r=1, L
           do 60 q=1, K
              do 60 p=1, H
                errc = fif(p,q,r)-fid(p,q,r)
                err=err+errc*errc
 60     continue
 	err=sqrt(err)/(H*K*L)
 	print *, 'Error: ',err
      
      return
      end
c
c ------------------------------------------------------------------------------
C
      real*8 Function rand()
C Random generator as used by D. Berkov.
      real*8 rnd
      common /rndnum/ rnd
      	rnd=rnd*997.0
      	rnd=rnd-int(rnd)
      	rand=rnd
      end
c
c ------------------------------------------------------------------------------
C
      Subroutine randseed(seed)
C Supply random generator with sane seed
C INPUT: real*8 seed ! seed for random generator, <>0!
C OUTPUT: VOID
      real*8 rnd, seed
      common /rndnum/ rnd
	rnd=seed
	return
      end
c
c ------------------------------------------------------------------------------
C
      Subroutine tr213(In, Out, HM,KM,LM)
C Transpose In(HM,KM,LM) to Out(KM,HM,LM). !!In and Out must be different!!
C INPUT: real*8 In(HM,KM,LM)
C 	integer HM,KM,LM
C OUTPUT: real*8 Out(KM,HM,LM)
      implicit none
      integer HM,KM,LM
      real*8 In(HM*KM*LM), Out(HM*KM*LM)
      integer i,j,k,p,q,r
c
	i=1
	do 1000 r=1, LM
	  j=(r-1)*KM*HM
	  do 1000 q=1, KM
	    k=q
	    do 1000 p=1, HM
	      Out(k+j) = In(i)
	      k=k+KM
	      i=i+1
1000	continue	        
      end
c
c ------------------------------------------------------------------------------
C
      Subroutine tr231(In, Out, HM,KM,LM)
C Transpose In(HM,KM,LM) to Out(KM,LM,HM). !!In and Out must be different!!
C INPUT: real*8 In(HM,KM,LM)
C 	integer HM,KM,LM
C OUTPUT: real*8 Out(KM,LM,HM)
      implicit none
      integer HM,KM,LM
      real*8 In(HM*KM*LM), Out(HM*KM*LM)
      integer i,j,k,p,q,r
c
	i=1
	do 1000 r=1, LM
	  k=(r-1)*KM
	  do 1000 q=1, KM
	    j=q
	    do 1000 p=1, HM
	      Out(k+j) = In(i)
	      j=j+(KM*LM)
	      i=i+1
1000	continue	        
      end
c
c ------------------------------------------------------------------------------
C
      Subroutine tr312(In, Out, HM,KM,LM)
C Transpose In(HM,KM,LM) to Out(LM,HM,KM). !!In and Out must be different!!
C INPUT: real*8 In(HM,KM,LM)
C 	integer HM,KM,LM
C OUTPUT: real*8 Out(LM,HM,KM)
      implicit none
      integer HM,KM,LM
      real*8 In(HM*KM*LM), Out(HM*KM*LM)
      integer i,j,k,p,q,r
c
	i=1
	do 1000 r=1, LM
	  do 1000 q=1, KM
	  j=(q-1)*LM*HM+r
	  k=j
	    do 1000 p=1, HM
	      Out(k) = In(i)
	      k=k+LM
	      i=i+1
1000	continue	        
      end
c
c ------------------------------------------------------------------------------
C
      Subroutine tr321(In, Out, HM,KM,LM)
C Transpose In(HM,KM,LM) to Out(LM,KM,HM). !!In and Out must be different!!
C INPUT: real*8 In(HM,KM,LM)
C 	integer HM,KM,LM
C OUTPUT: real*8 Out(LM,KM,HM)
      implicit none
      integer HM,KM,LM
      real*8 In(HM*KM*LM), Out(HM*KM*LM)
      integer i,j,k,p,q,r
c
	i=1
	do 1000 r=1, LM
	  do 1000 q=1, KM
	    k=(q-1)*LM+r
	    j=k
	    do 1000 p=1, HM
	      Out(j) = In(i)
	      j=j+LM*KM
	      i=i+1
1000	continue	        
      end
c
c ------------------------------------------------------------------------------
C
      real*8 function diff(v1, v2, siz)
C Check that v1 and v2 are equal: return sum of all differences squared      
C INPUT: real*8 v1(siz), v2(siz)
C	integer siz
C OUTPUT: function value is sum(i=1..siz, (v1(i)-v2(i))**2)
      implicit none
      integer siz
      real*8 v1(siz),v2(siz)
      integer i
      real*8 errc, err
c
	err=0
	do 100 i=1, siz
	  errc=v1(i)-v2(i)
	  err=err+errc*errc
100	continue	      
c
	diff=err
      end
c
c ------------------------------------------------------------------------------
C
      Subroutine CVFFTTIM
C Measure FFT time as function of input length
      implicit none
      include 'common.inc'
      real*8 wsav(NM*4), seq(NM*4), seqf(NM*4)
      real*8 rand, err, tim, timer, delay, diff, stp
      external rand, timer, diff
      integer Ns, Ne, NM4, j, k, rep, vb
c
c declare these arrays common so that they dont cost stack space:
      common /cvdbgCommon/ wsav, seq, seqf
c
	NM4 = NM*4
 100	continue
	print *, 'CVFFTTIM: Measure fft time as func. of input size'
	print *, 'Start size (<',4*NM,'):'
	read(*,*,ERR=100) Ns
	print *, 'End size (<',4*NM,'):'
	read(*,*,ERR=100) Ne
	if(Ns.lt.0.or.Ne.lt.0.or.Ns.gt.NM4.or.Ne.gt.NM4) go to 100
	print *, 'approx. time/step:'
	read(*,*,ERR=100) stp
	print *, 'Verbosity (0/1):'
	read(*,*,ERR=100) vb
	
	call RandSeed(3.14D0)
	do 150 k=1, Ne
	  seq(k)=rand()
 150	continue
 
	do 200, k=Ns, Ne
	  call rffti(k,wsav)

	  tim=timer()
	  rep=0
 400	  continue	  
	    do 300 j=1, k
	      seqf(j)=seq(j)
 300	    continue
	    call rfftf(k, seqf, wsav)
	    call rfftb(k, seqf, wsav)
	    rep=rep+1
	    delay = timer()-tim
	  if( delay.lt.stp) go to 400
	  do 350 j=1, k
	      seqf(j)=seqf(j)/k
 350	    continue
	  err=diff(seq,seqf,k)
	  if(vb.gt.0) then
	    print *, rep, ' cycles averaged to ',delay/rep,' ffts/s',
     &			' of size',k	    
	    print *, 'Error: ', err
	  else
	    print *, k, delay/rep, rep, err
	  endif
 200	continue

 	return
       end
C===============================================================================
