C---.----1----.----2----.----3----.----4----.----5----.----6----.----7----.----
C 
C cvlvd.f: direct convolution
C	This is only to check that other, more sophisticated methods work.
C
C Authors: P. Lockey and K. Ramstoeck, 17. Oct. 1995
C
C==============================================================================
      Subroutine CV3DDInit(din, K,L,M, KM,LM, dout, sz)
C Take the array din(1:NM, 1:M) and prepare it for convolution. 
C din(*) is used as din(1:N, 1:M) in the caller. dout is allocated as
C dout(1:sz) by the caller (must be N*M at least)
C All that is done here is copy this into a contiguous array
C upon return, sz contains the number of reals actually used in dout, or
C -1 if an error occured (sz too small)
      implicit none
      integer K,L,M, KM,LM, sz
      real*8 din(1:KM, 1:LM, 1:M), dout(1:sz)
      integer p,q,r,kl
c
c
        if(sz.lt.K*L*M) then
          sz=-1
          return
        endif
c
c 1. Step: Copy din to dout
	kl=K*L
        do 100 r=1, M
           do 100 q=1, L
              do 100 p=1, K
                dout((r-1)*kl+(q-1)*K+p)=din(p,q,r)
 100    continue
c
        sz = K*L*M
c
        return
      end
c
c
c
      Subroutine CV3DD(d, K,L,M, KM,LM, if, dout)
C Take the data in d (declared d(1:KM, 1:LM, 1:M), used as d(1:K,1:L,1:M)
C and convolute it with the function in if. The result will be returned
C in dout(1:K,1:L,1:M). dout and d can be the same arrays.
C if will not be changed, so it can be used several times.
C if has to be prepared with CV3DDInit
      implicit none
      integer K,L,M, KM,LM
      real*8 d(KM, LM, M), if(0:K-1, 0:L-1, 0:M-1), dout(KM, LM, M)
      real*8 dtmp(KM, LM, M), tmp
      integer p,q,r, h,i,j, ar, aq
c
      do 10 j=1, M
         do 10 i=1, L
            do 10 h=1, K
               tmp=0
               do 20 r=1, M
               	  ar=abs(r-j)
                  do 20 q=1, L
                     aq=abs(q-i)
                     do 20 p=1, K
                        tmp=tmp+d(p,q,r)*if(abs(p-h),aq,ar)
 20            continue
               dtmp(h,i,j)=tmp
 10   continue
c
c
      do 30 j=1, M
         do 30 i=1, L
            do 30 h=1, K
               dout(h,i,j)=dtmp(h,i,j)
 30   continue
c
c
        return
      end
c
c

      Subroutine CV2DDInit(din, N, M, NM, dout, sz)
C Take the array din(1:NM, 1:M) and prepare it for convolution. 
C din(*) is used as din(1:N, 1:M) in the caller. dout is allocated as
C dout(1:sz) by the caller (must be N*M at least)
C All that is done here is copy this into a contiguous array
C upon return, sz contains the number of reals actually used in dout, or
C -1 if an error occured (sz too small)
      implicit none
      integer N, M, NM, sz
      real*8 din(1:NM, 1:M), dout(1:sz)
      integer i,j
c
c
        if(sz.lt.N*M) then
          sz=-1
          return
        endif
c
c 1. Step: Copy din to dout
        do 100 j=1, M
           do 100 i=1, N
              dout((j-1)*N+i)=din(i,j)
 100    continue
c
        sz = N*M
c
        return
      end
c
c
c
      Subroutine CV2DD(d, N, M, NM, if, dout)
C Take the data in d (declared d(1:NM, 1:M), used as d(1:N,1:M)) and
C convolute it with the function in if. The result will be returned in
C dout, which can be the same array as d.
C if will not be changed, so it can be used several times.
C if has to be prepared with CV2DDInit
      implicit none
      integer N,M,NM
      real*8 d(1:NM, 1:M), if(0:N-1, 0:M-1), dout(1:NM,1:M)
      real*8 dtmp(N,M), tmp
      integer i,j, h,k, ajk
c
      do 10 j=1, M
         do 10 i=1, N
            tmp=0
            do 20 k=1, M
               ajk=abs(j-k)
               do 20 h=1, N
                  tmp=tmp+d(h,k)*if(abs(i-h),ajk)
 20         continue
            dtmp(i,j)=tmp
 10   continue
c
      do 30 j=1, M
         do 30 i=1, N
            dout(i,j)=dtmp(i,j)
 30   continue
c
        return
      end
c
c
c
      Subroutine CV1DDInit(din, N, M, NM, dout, sz)
C Take the array din(1:NM, 1:M) and prepare it for convolution. 
C din(*) is used as din(1:N, 1:M) in the caller. dout is allocated as
C dout(1:sz) by the caller (must be N*M at least)
C All that is done here is copy this into a contiguous array
C upon return, sz contains the number of reals actually used in dout, or
C -1 if an error occured (sz too small)
      implicit none
      integer N, M, NM, sz
      real*8 din(1:NM, 1:M), dout(1:sz)
      integer i,j
c
c
        if(sz.lt.N*M) then
          sz=-1
          return
        endif
c
c 1. Step: Copy din to dout
        do 100 j=1, M
           do 100 i=1, N
              dout((j-1)*N+i)=din(i,j)
 100    continue
c
        sz = N*M
c
        return
      end
c
c
c
      Subroutine CV1DD(d, N, M, NM, if, Mif)
C Take the data in d (declared d(1:NM, 1:M), used as d(1:N,1:M)) and
C convolute it with the function in if. The result will be returned in d.
C if will not be changed, so it can be used several times.
C if has to be prepared with CV2DDInit
      implicit none
      integer N,M,NM,Mif
      real*8 d(1:NM, 0:Mif-1), if(0:N-1, 1:M)
      real*8 dtmp(1:N,1:Mif), tmp
      integer i,j, h,k
c
      do 10 j=1, Mif
         do 10 i=1, N
            tmp=0
            k=mod(j-1, M)
            do 20 h=1, N
               tmp=tmp+d(h,k)*if(abs(i-h),j)
 20         continue
            dtmp(i,j)=tmp
 10   continue
c
      do 30 j=1, Mif
         do 30 i=1, N
            d(i,j-1)=dtmp(i,j)
 30   continue
c
        return
      end
