C---.----1----.----2----.----3----.----4----.----5----.----6----.----7----.----
C 
C cvlvp.f: parallel convolution
C	This code uses the Netlib FFT-Routines. Work is distributed onto
C       N processors, each of them having a copy of the total data.
C       Using a 'WorkSplit' function, the time consuming parts are
C       split among the processors. Later, a Subroutine Merge is called
C       to collect the results.
C
C DEFICIENCIES/BUGS:
C  The synopsis of the 1D-routines is a bit weird, and they are so highly
C  specialized they might not be of use to anybody else, but then ..
C
C Authors: P. Lockey and K. Ramstoeck, 17. Oct. 1995
C
C==============================================================================
      Subroutine CV3DInit(din, H,K,L, HM,KM, dout, sz)
C Take the array din(HM,KM,L) and prepare it for convolution using FFT.
C din(*) is used as din(H,K,L) in the caller. dout is allocated as
C dout(1:sz) by the caller (sz must be 8*H*K*L at least)
C upon return, sz contains the number of reals actually used in dout, or
C -1 if an error occured (sz too small)
C NOTE:
C * This works for NON-periodic functions, as all maters of zero-padding
C   are taken care of.
C * The transpose-operations are (hopefully) arranged as to cause minimum
C   cache problems (Tested on my HP 9000/720 by KR)
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer H,K,L, HM,KM, sz
      real*8 din(HM, KM, L), dout(sz)
      integer i,j,f,g, ad,p,q,r, LM,H2,K2,L2
      integer had(nmax),kad(nmax),lad(nmax)
      real*8 dtmp1(H*2*K*2*L*2),dtmp2(H*2*K*2*L), scal
      integer KL,KLsz,KLlo,KLhi,HK2,HK2sz,HK2lo,HK2hi, minsz
      integer HL4,HL4sz,HL4lo,HL4hi,L2K
      integer w1sz, w2sz, w3sz
      real*8 wsave1(2*NMax+15), wsave2(2*NMax+15), wsave3(2*NMax+15)
      common /cvlvpNL/ wsave1,wsave2,wsave3, w1sz,w2sz,w3sz
c
c Initialization stuff
c Get 'Nice sizes' for FFT, meaning N2.GE.2*N and N2=2^A*3^B*5^C; same for M2
        Call CVSizes3D(H,K,L, H2,K2,L2, sz)
        LM=L
        L2K=L2*K
        scal=1.0D0/(H2*K2*L2)		! Assume integer*4 at least ...
        minsz=(H+1)*(K+1)*(L+1)		! Minimum size for dout
c
        if(sz.lt.minsz
     &       .or. H2.gt.NMax .or. K2.gt.NMax .or. L2.gt.NMax) then
          sz=-minsz
          return
        endif
c
        call rffti(H2, wsave1)
        w1sz=H2
        call rffti(K2, wsave2)
        w2sz=K2
        call rffti(L2, wsave3)
        w3sz=L2
c
c Get parallel partitioning info.
	KL=K*L
	HK2=H*K*2
	HL4=H*L*4
        call pSplit(KL,KLsz,KLlo,KLhi)
        call pSplit(HK2,HK2sz,HK2lo,HK2hi)
        call pSplit(HL4,HL4sz,HL4lo,HL4hi)
c
c 1. Step: Copy din to tmp and wrap around. Assumes H2=2*H
	ad=1
        do 100 r=1, L
          do 100 q=1, K
            do 110 p=1, H
	      dtmp1(ad)=din(p,q,r)
	      ad=ad+1
 110        continue
 	    dtmp1(ad)=0
 	    ad=ad+1
            do 120 p=H, 2,-1
	      dtmp1(ad)=din(p,q,r)
	      ad=ad+1
 120        continue
 100    continue
c        call PrtArray3D(dtmp1, H2, K, L, H2,K, 'CV3DInit-input data:')
c
c 2. Step: Perform K*L FFT's of length H2
        do 200 j=KLlo+1, KLhi+1
           call rfftf(H2, dtmp1((j-1)*H2+1), wsave1)
 200    continue
c        call PrtArray3D(dtmp1, H2, K, L, H2,K, '1st FFT:')
c	call PMerge(...)
c
c 3. Step: Transpose of type 321: dtmp1(H2,K,L)-> dtmp2(L2,K,H2) and
c   wrap around. Assumes L2=2*L
	ad=1
	do 300 r=1, L
	  do 300 q=1, K
	    j=(q-1)*L2+r
	    do 300 p=1, H2
	      dtmp2(j) = dtmp1(ad)
	      j=j+L2K
	      ad=ad+1
 300	continue
	ad=1
 	do 310 j=1, HK2
	  dtmp2(ad+L)=0.0D0
	  ad=ad+1
	  f=ad+L2-2
 	  do 320 i=2, L
 	    dtmp2(f) = dtmp2(ad)
 	    ad=ad+1
 	    f=f-1
 320 	  continue
 	  ad=ad+L
 310	  continue
 
c	call PrtArray3D(dtmp2, L2, K, H2, L2,K, '1st Transpose:')
c
c 4. Step: Perform 2*H*K FFT's of length L2
	do 400 j=HK2lo+1, HK2hi+1
	  call rfftf(L2, dtmp2((j-1)*L2+1), wsave3)
 400	continue
c        call PrtArray3D(dtmp2, L2, K, H2, L2,K, '2.nd FFT:')
c 	call PMerge(...)
c
c 5. Step: Transpose of type 213: dtmp2(L2,K,H2)-> dtmp1(K2,L2,H2)
	ad=1
	do 500 r=1, H2
	  i=(r-1)*K2*L2
	  do 500 q=1, K
	    j=0
	    do 500 p=1, L2
	      dtmp1(i+j+q) = dtmp2(ad)
	      ad=ad+1
	      j=j+K2
 500	continue
	ad=1
 	do 510 j=1, HL4
	  dtmp1(ad+K)=0.0D0
	  ad=ad+1
	  f=ad+K2-2
 	  do 520 i=2, K
 	    dtmp1(f) = dtmp1(ad)
 	    f=f-1
 	    ad=ad+1
 520 	  continue
	  ad=ad+K
 510	continue
c        call PrtArray3D(dtmp1, K2,L2,H2, K2,L2, '2nd Transpose:')
c
c 6. Step: Perform 4*H*L FFT's of length K2
	do 600 j=HL4lo+1, HL4hi+1
	  call rfftf(K2, dtmp1((j-1)*K2+1), wsave2)
 600	continue
c        call PrtArray3D(dtmp1, K2,L2,H2, K2,L2, '3rd FFT:')
c 	call PMerge(...)
c
c 7. Step: Copy real components of dtmp into dout
	had(1)=1
	do 700 r=2, H2/2+1
	  had(r)=(r-1)*2
700	continue

	lad(1)=1
	do 710 q=2, L2/2+1
	  lad(q)=(q-1)*2
710	continue

	kad(1)=1
	do 720 p=2, K2/2+1
	  kad(p)=(p-1)*2
720	continue

	ad=1
	do 730 r=1, H2/2+1
	  i=(had(r)-1)*K2*L2
	  do 730 q=1, L2/2+1
	    j=i+(lad(q)-1)*K2
	    do 730 p=1, K2/2+1
	      dout(ad) = dtmp1(j+kad(p))*scal
	      ad=ad+1
730	continue
c        call PrtArray3D(dout, K2/2+1,L2/2+1,H2/2+1,
c     & 		K2/2+1,L2/2+1, 'Output:')
c
        sz = (H2/2+1)*(K2/2+1)*(L2/2+1)
c
        return
      end
c CV3DInit end.
c
c
      Subroutine CV3D(d, H,K,L, HM,KM, if, dout, h0,k0,l0,
     &     HMout,KMOut, scr,lscr)
C Take the data in d (declared d(HM,KM,L), used as d(H,K,L)) and
C convolute it with the function in if using FFT. The result will be 
C returned in dout(h0:h0+H-1,k0:k0+K-1,l0:l0+l-1). dout is declared by
C the caller as dout(1:HMout,1:KMOut,1:l0+l-1) at least. if will not be
C changed, so it can be used several times. if has to be prepared with
C CV2DInit. scr is a scratch place of at least size 12*H*K*L, if it is
C to be used as signaled by lscr=1. If lscr=0, scr might be a dummy and
C the scratch space will be taken from the stack.
      implicit none
      integer H,K,L, HM,KM, h0,k0,l0, HMOut,KMOut, lscr
      real*8 d(HM,KM,L), if(8*HM*KM*L) ! ???
      real*8 dout(1:HMout,1:KMOut,1:l0+L-1), scr(0:12*H*K*L*lscr)
      integer sz
      real*8 dtmp1(0:8*H*K*L*(1-lscr)),dtmp2(0:4*H*K*L*(1-lscr))
c
      sz=8*H*K*L ! ???
        if(lscr.eq.1)
     &    call CV3Dpriv(d, H,K,L, HM,KM, if, dout, h0,k0,l0,
     &     HMout,KMOut, scr, scr(sz))
        if(lscr.eq.0)
     &    call CV3Dpriv(d, H,K,L, HM,KM, if, dout, h0,k0,l0,
     &     HMout,KMOut, dtmp1, dtmp2)
        return
      end
c
c
c
      Subroutine CV3Dpriv(d, H,K,L, HM,KM, if, dout, h0,k0,l0,
     &     HMout,KMOut, dtmp1, dtmp2)
C Private implementation of CV3D, don't call directly
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer H,K,L, HM,KM, h0,k0,l0, HMOut,KMOut
      real*8 d(HM,KM,L), if(8*HM*KM*L) ! ???
      real*8 dout(1:HMout,1:KMOut,1:l0+L)
      real*8 dtmp1(8*H*K*L),dtmp2(4*H*K*L)
      integer KL,KLsz,KLlo,KLhi,HK2,HK2sz,HK2lo,HK2hi
      integer HL4,HL4sz,HL4lo,HL4hi, H2,K2,L2,KL1,sz,lm,L2K
      integer r,q,p,i,j,ad,f,g
      integer had(nmax),kad(nmax),lad(nmax)
      integer w1sz, w2sz, w3sz
      real*8 wsave1(2*NMax+15), wsave2(2*NMax+15), wsave3(2*NMax+15)
      common /cvlvpNL/ wsave1,wsave2,wsave3, w1sz,w2sz,w3sz
c
c
c Initialization stuff
c Get 'Nice sizes' for FFT
        Call CVSizes3D(H,K,L, H2,K2,L2, sz)
        LM=L
        L2K=L2*K
c
	if(w1sz.ne.H2) then
	  call rffti(H2,wsave1)
	  w1sz=H2
	endif
	if(w2sz.ne.K2) then
	  call rffti(K2,wsave2)
	  w2sz=K2
	endif
	if(w3sz.ne.L2) then
	  call rffti(L2,wsave3)
	  w3sz=L2
	endif
c
c Get parallel partitioning info.
	KL=K*L
	HK2=H*K*2
	HL4=H*L*4
        call pSplit(KL,KLsz,KLlo,KLhi)
        call pSplit(HK2,HK2sz,HK2lo,HK2hi)
        call pSplit(HL4,HL4sz,HL4lo,HL4hi)

c
c 1. Step: Copy d to tmp an zero padd
	ad=1
        do 100 r=1, L
          do 100 q=1, K
            do 110 p=1, H
	      dtmp1(ad)=d(p,q,r)
	      ad=ad+1
 110	    continue
            do 120 p=1, H
	      dtmp1(ad)=0.0D0
	      ad=ad+1
 120	    continue
 100    continue
c        call PrtArray3D(dtmp1, H2, K, L, H2,K, 'CV3D-Input Data:')
c
c 2. Step: Perform K*L FFT's of length H2
        do 200 j=KLlo+1, KLhi+1
           call rfftf(H2, dtmp1((j-1)*H2+1), wsave1)
 200    continue
c        call PrtArray3D(dtmp1, H2, K, L, H2,K, '1st FFT:')
c	call PMerge(...)
c
c 3. Step: Transpose of type 321: dtmp1(H2,K,L)-> dtmp2(L2,K,H2)
	ad=1
	do 300 r=1, L
	  do 300 q=1, K
	    j=(q-1)*L2+r
	    g=(q-1)*L2+r+L
	    do 300 p=1, H2
	      dtmp2(j) = dtmp1(ad)
	      dtmp2(g) = 0.0D0
	      j=j+L2K
	      g=g+L2K
	      ad=ad+1
 300	continue
c        call PrtArray3D(dtmp2, L2, K, H2, L2,K, '1st Transpose:')
c
c 4. Step: Perform 2*H*K FFT's of length L2
	do 400 j=HK2lo+1, HK2hi+1
	  call rfftf(L2, dtmp2((j-1)*L2+1), wsave3)
 400	continue
c        call PrtArray3D(dtmp2, L2, K, H2, L2,K, '2.nd FFTe')
c 	call PMerge(...)
c
c 5. Step: Transpose of type 213: dtmp2(L2,K,H2)-> dtmp1(K2,L2,H2)
	ad=1
	do 500 r=1, H2
	  i=(r-1)*K2*L2
	  do 500 q=1, K
	    j=0
	    f=K
	    do 500 p=1, L2
	      dtmp1(i+j+q) = dtmp2(ad)
	      dtmp1(i+f+q) = 0.0D0
	      ad=ad+1
	      j=j+K2
	      f=f+K2
 500	continue
c        call PrtArray3D(dtmp1, K2,L2,H2, K2,L2, '2nd Transpose:')
c
c 6. Step: Perform 4*H*L FFT's of length K2
	do 600 j=HL4lo+1, HL4hi+1
	  call rfftf(K2, dtmp1((j-1)*K2+1), wsave2)
 600	continue
c        call PrtArray3D(dtmp1, K2,L2,H2, K2,L2, '3rd FFT:')
c 	call PMerge(...)
c
c 7. Step: Convolution
        do 700 i=1, H2
           had(i)=i/2+1
 700    continue

        do 710 i=1, L2
           lad(i)=i/2+1
 710    continue

        do 720 i=1, K2
           kad(i)=i/2+1
 720    continue
 
	ad=1
	KL1=(K2/2+1)*(L2/2+1)
	do 730 r=1, H2
	  i=(had(r)-1)*KL1
	  do 730 q=1, L2
	    j=i+(lad(q)-1)*(K2/2+1)
	    do 730 p=1, K2
	      dtmp1(ad) = dtmp1(ad)*if(j+kad(p))
	      ad=ad+1
730	continue
c        call PrtArray3D(dtmp1, K2,L2,H2, K2,L2, 'Convolution:')
c
c 8. Step: Perform 4*H*L FFT's of length K2
	do 800 j=HL4lo+1, HL4hi+1
	  call rfftb(K2, dtmp1((j-1)*K2+1), wsave2)
 800	continue
c        call PrtArray3D(dtmp1, K2,L2,H2, K2,L2, '1st FFT back:')
c 	call PMerge(...)
c
c 9. Step: Transpose of type 213: dtmp1(K2,L2,H2)-> dtmp2(L2,K,H2)
	ad=1
	do 900 r=1, H2
	  i=(r-1)*K2*L2
	  do 900 q=1, K
	    j=0
	    do 900 p=1, L2
	      dtmp2(ad) = dtmp1(i+j+q)
	      ad=ad+1
	      j=j+K2
 900	continue
c        call PrtArray3D(dtmp2, L2,K,H2,L2,K, '1st Transpose back:')
c
c 10. Step: Perform 2*H*K FFT's of length L2
	do 1000 j=HK2lo+1, HK2hi+1
	  call rfftb(L2, dtmp2((j-1)*L2+1), wsave3)
 1000	continue
c        call PrtArray3D(dtmp2, L2,K,H2,L2,K, '2nd FFT back:')
c 	call PMerge(...)
c
c 11. Step: Transpose of type 321: dtmp2(L2,K,H2) -> dtmp1(H2,K,L)
	ad=1
	do 1100 r=1, L
	  do 1100 q=1, K
	    j=(q-1)*L2+r
	    do 1100 p=1, H2
	      dtmp1(ad) = dtmp2(j) 
	      j=j+L2K
	      ad=ad+1
 1100	continue
c        call PrtArray3D(dtmp1, H2,K,L,H2,K, '2nd Transpose back:')
c
c 12. Step: Perform K*L FFT's of length H2
        do 1200 j=KLlo+1, KLhi+1
           call rfftb(H2, dtmp1((j-1)*H2+1), wsave1)
 1200    continue
c        call PrtArray3D(dtmp1, H2,K,L,H2,K, '3rd FFT back:')
c	call PMerge(...)
c
c 13. Step: Copy result into dout
	ad=dout(H,K,L)
	ad=1
        do 1300 r=l0, L+l0-1
          do 1300 q=k0, K+k0-1
            do 1310 p=h0, H+h0-1
	      dout(p,q,r)=dtmp1(ad)
	      ad=ad+1
 1310	    continue
 	    ad=ad+H
 1300    continue
c
c	call PrtArray3D(dout, HMOut, KMOut, L, HMOut, KMOut, 'dout:')
        return
      end
c CV3D End
c
c
      Subroutine CV2DInit(din, N, M, NM, dout, sz)
C Take the array din(1:NM, 1:M) and prepare it for convolution using FFT.
C din(*) is used as din(1:N, 1:M) in the caller. dout is allocated as
C dout(1:sz) by the caller (sz must be 4(N+1)(M+1) at least)
C upon return, sz contains the number of reals actually used in dout, or
C -1 if an error occured (sz too small)
C NOTE: This works for NON-periodic functions, as all maters of zero-padding
C are taken care of.
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer N, M, NM, sz
      real*8 din(1:NM, 1:M), dout(sz)
      integer i,j, N2,M2, naddr(nmax),maddr(nmax)
      real*8 dtmp(NMax*NMax/2)
      integer n2sz,n2lo,n2hi,msz,mlo,mhi, minsz
      integer w1sz, w2sz, w3sz
      real*8 wsave1(2*NMax+15), wsave2(2*NMax+15), wsave3(2*NMax+15)
      common /cvlvpNL/ wsave1,wsave2,wsave3, w1sz,w2sz,w3sz
c
C Initialization stuff
C Get 'Nice sizes' for FFT, meaning N2.GE.2*N and N2=2^A*3^B*5^C; same for M2
        Call CVSizes(N,M,N2,M2, sz)
        minsz=(N+1)*(M+1)
c
        if(sz.lt.minsz .or. N2.gt.NMax .or. M2.gt.NMax) then
          sz=-minsz
          return
        endif
c
c Get parallel partitioning info.
        call pSplit(n2,n2sz,n2lo,n2hi)
        call pSplit(m,msz,mlo,mhi)
c
c 1. Step: Zero out dout and dtmp array:
        do 100 j=1, N2*M
           dtmp(j)=0.0D0
 100    continue
        do 105 j=1, N2*M2
           dout(j)=0.0D0
 105    continue
c
c 2. Step: copy din to dtmp
c 2.1 Copy first part
        do 110 j=1, M
           do 110 i=1, N
              dtmp((j-1)*N2+i)=din(i,j)
 110    continue
c 2.2 Copy wrapped-around part
        do 115 j=1, M
           do 115 i=2, N
              dtmp((j-1)*N2+N2-i+2)=din(i,j)
 115    continue
cc        call PrtArray(dtmp, N2, M, N2, 'Init-dtmp:')  ! <-
c
c 3. Step: Initialize WSAVE arrays
        call rffti(N2, wsave1)
        w1sz=N2
        call rffti(M2, wsave2)
        w2sz=M2
c
c 4. Step: fft in 1.st direction
c        write(*,*) '(first fftf) N2 = ',n2
        do 120 j=mlo+1, mhi+1
           call rfftf(N2, dtmp((j-1)*N2+1), wsave1)
 120    continue
cc        call PrtArray(dtmp, N2, M, N2, 'Init-1.st fft:')  ! <-
c
c 5. Step: Merge the fft'ed data
        call pMerge(dtmp,M,N2)
c
c 6. Step: Transpose dtmp to dout
c 6.1 Transpose first part
        do 130 j=1, M
           do 130 i=n2lo+1, n2hi+1
              dout((i-1)*M2+j) = dtmp((j-1)*N2+i)
 130    continue
c 6.2 Transpose wrapped-around part
        do 135 j=2, M
           do 135 i=n2lo+1, n2hi+1
              dout((i-1)*M2+M2-j+2) = dtmp((j-1)*N2+i)
 135    continue
cc        call PrtArray(dout, M2, N2, M2, 'Init-transpose')  ! <-
c
c 7. Step fft 2nd direction
c        write(*,*) '(2nd fftf) N2 = ',n2
        do 140 i=n2lo+1, n2hi+1
           call rfftf(M2, dout((i-1)*M2+1), wsave2)
 140    continue
cc        call PrtArray(dout, M2, N2, M2, 'Init-2nd FFT')  ! <-
c
c 8. Step: Merge results
        call pMerge(dout,N2,M2)
c 9. Step: Compress out imaginary parts (0 due to symmetry). This can be
c done in place.
        naddr(1)=1
        do 145 i=2, N2/2+1
           naddr(i)=(i-1)*2
 145    continue
c        write(*,*) (naddr(i),i=1,N2/2+1)
        maddr(1)=1
        do 146 j=2, M2/2+1
           maddr(j)=(j-1)*2
 146    continue
c        write(*,*) (maddr(j),j=1,M2/2+1)
        do 150 i=1, N2/2+1 
           do 150 j=1, M2/2+1
              dout((i-1)*(M2/2+1)+j)=dout((naddr(i)-1)*M2+maddr(j))
 150    continue
c
        sz = (N2/2+1)*(M2/2+1)
c        sz = N2*M2         ! If imaginary parts are not stripped
c
        return
      end
c CV2DInit end.
c
c
      Subroutine CV2D(d,N,M,NM,if,dout,i0,j0,NMout,scr,lscr)
C Take the data in d (declared d(1:NM, 1:M), used as d(1:N,1:M)) and
C convolute it with the function in if using FFT. The result will be 
C returned in dout(i0:i0+N-1,j0:j0+M-1). dout is declared by the caller
C as dout(1:NMout,1:M+j0-1) at least. if will not be changed, so it can
C be used several times. if has to be prepared with CV2DInit.
C scr is a scratch place of at least size 6*N*M, if it is
C to be used as signaled by lscr=1. If lscr=0, scr might be a dummy
      implicit none
      integer N,M,NM,i0,j0,NMout,lscr
      real*8 d(1:NM, 1:M), if(4*(N+2)*(M+2))
      real*8 dout(1:NMout,1:M+j0-1), scr(0:6*N*M*lscr)
      integer sz
      real*8 dtmp1(0:2*N*M*(1-lscr)),dtmp(0:4*N*M*(1-lscr))
c
      sz=2*N*M+1
        if(lscr.eq.1)
     &    call CV2Dpriv(d,N,M,NM,if,dout,i0,j0,NMout,scr(0),scr(sz))
        if(lscr.eq.0)
     &    call CV2Dpriv(d,N,M,NM,if,dout,i0,j0,NMout,dtmp1,dtmp)
        return
      end
c
c
c
      Subroutine CV2Dpriv(d,N,M,NM,if,dout,i0,j0,NMout,dtmp1,dtmp)
C Private implementation of CV2D, don't call directly
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer N,M,NM,i0,j0,NMOut
      real*8 d(1:NM, 1:M), if(4*N*M)
      real*8 dout(1:NMout,1:M+j0-1)
      real*8 dtmp1(2*N*M), dtmp(4*N*M), scal
      integer i,j, N2,M2, adr,adr2, mdim
      integer naddr(nmax),maddr(nmax)
      integer msz,mlo,mhi,n2sz,n2lo,n2hi
      integer w1sz, w2sz, w3sz
      real*8 wsave1(2*NMax+15), wsave2(2*NMax+15), wsave3(2*NMax+15)
      common /cvlvpNL/ wsave1,wsave2,wsave3, w1sz,w2sz,w3sz
c
C Initialization:
        Call CVSizes(N,M,N2,M2,-1)
        do 90 i=1, N2
           naddr(i)=i/2+1
 90     continue
c        write(*,*) (naddr(i),i=1,N2)
        do 95 j=1, M2
           maddr(j)=j/2+1
 95     continue
c        write(*,*) (maddr(j),j=1,M2)
c
c Get partitioning info
        call pSplit(m,msz,mlo,mhi)
        call pSplit(n2,n2sz,n2lo,n2hi)
c
c 1. Step: Initialize tmp storage to zero
        do 100 i=1, N2*M
           dtmp1(i)=0.0D0
 100    continue
        do 110 i=1, N2*M2
           dtmp(i)=0.0D0
 110    continue
c 1.1 Check if w1sz and N2 match
        if (w1sz.ne.N2) then
           call rffti(N2, wsave1)
           w1sz=N2
        endif
c 1.2 Check if w2sz and M2 match
        if (w2sz.ne.M2) then
           call rffti(M2, wsave2)
           w2sz=M2
        endif
c
c 2. Step: copy d to dtmp1
        do 120 j=1, M
           do 120 i=1, N
              dtmp1((j-1)*N2+i)=d(i,j)
 120    continue
cc       call PrtArray(dtmp1, N2, M, N2, 'cv2d-d inserted')  ! <-
c
c 3. Step: FFT dtmp1
        do 130 j=mlo+1, mhi+1
           call rfftf(N2, dtmp1((j-1)*N2+1), wsave1)
 130    continue
cc        call PrtArray(dtmp1, N2, M, N2, 'cv2d-1st dim FFT fwd')  ! <-
c
c 4. Step: Transpose dtmp1 to dtmp
c 4.1 Merge the results from the FFT first
        call pMerge(dtmp1,M,N2)
c 4.2 Transpose first part
        do 140 j=1, M
           do 140 i=n2lo+1, n2hi+1
              dtmp((i-1)*M2+j) = dtmp1((j-1)*N2+i)
 140    continue
cc        call PrtArray(dtmp, M2, N2, M2, 'cv2d-Transposed')  ! <-
c
c 5. Step: FFT dtmp in 2.nd direction
        do 150 i=n2lo+1, n2hi+1
           call rfftf(M2, dtmp((i-1)*M2+1), wsave2)
 150    continue
cc        call PrtArray(dtmp, M2, N2, M2, 'cv2d-2nd dim FFT fwd')  ! <-
c
c 6. Step: Multiply arrays
        mdim=M2/2+1
        do 160 i=n2lo+1, n2hi+1
           do 160 j=1, M2
              adr=(i-1)*M2+j
              adr2 = (naddr(i)-1)*Mdim+maddr(j)
              dtmp(adr) = dtmp(adr)*if(adr2)
 160    continue
cc       call PrtArray(dtmp, M2, N2, M2, 'cv2d-Multiply')  ! <-
c
c 7. Step: FFT dtmp 2.nd direction back
        do 170 i=n2lo+1, n2hi+1
           call rfftb(M2, dtmp((i-1)*M2+1), wsave2)
 170    continue
cc       call PrtArray(dtmp, M2, N2, M2, 'cv2d-2nd dim FFT back')  ! <-
c
c 8. Step: Transpose dtmp to dtmp1
c 8.1 Merge
        call pMerge(dtmp,n2,m2)
c 8.2 Transpose
        do 180 j=mlo+1, mhi+1
           do 180 i=1, N2
              dtmp1((j-1)*N2+i) = dtmp((i-1)*M2+j)
 180    continue
cc       call PrtArray(dtmp1, N2, M, N2, 'cv2d-Transpose back')  ! <-
c
c 9. Step: FFT 1. Dimension back
        do 190 j=mlo+1, mhi+1
           call rfftb(N2, dtmp1((j-1)*N2+1), wsave1)
 190    continue
cc       call PrtArray(dtmp1, N2, M, N2, 'cv2d-1st dim FFT back')  ! <-
c
c 10. Step: Copy relevant elements of dtmp1 to dout
        call pMerge(dtmp1,m,n2)
        scal=N2*M2
        scal=1/scal
        do 200 j=1, M
           do 200 i=1, N
              dout(i+i0-1,j+j0-1) = dtmp1((j-1)*N2+i)*scal
 200    continue
c
c Uff. 11. Step: Go Home.
        return
      end
c CV2D End
c
c
      Subroutine CV1DInit(din,N,M,NM,dout,sz)
C Take the array din(1:NM, 1:M) and prepare it for convolution using FFT.
C Convolution is only performed in the 1st dimension.
C din(*) is used as din(1:N, 1:M) in the caller. dout is allocated as
C dout(1:sz) by the caller (must be 2*N*M at least)
C upon return, sz contains the number of reals actually used in dout, or
C -1 if an error occured (sz too small)
C Note: This works for NON-periodic functions, as all maters of zero-padding
C are taken care of
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer N, M, NM, sz
      real*8 din(1:NM, 1:M), dout(sz)
      integer i,j, N2, naddr(NMax)
      integer WSlots, WSlotSize, Slot, Saddr
      parameter (WSlots=2, WSlotSize=2*NMax+15)  ! # Wsave Arr. + Size in reals
      integer wsz(WSlots), Wused(WSlots)        ! # N for Wsav + last-used flag
      real*8 wsave(WSlotSize*WSlots)            ! Wsave arrays themselves
      integer msz, mlo, mhi, minsz
      common /CV1Dcommon/ Wsz, wsave, Wused     ! Declare common
      data wsz/-1,-1/, Wused/0,0/               ! Initialize sane
c
c Initialization+sanity checks
        Call CVSizes(N,1,N2,1,sz)
        minsz=(N+1)*M
        if(sz.lt.minsz .or. N2.gt.NMax) then
          sz=-minsz
          return
        endif
        Slot=1
        if( Wused(Slot).eq.1 ) Slot=2
        Wused(1)=0
        Wused(2)=0
        Wused(slot)=1
        Wsz(slot)=N2
        Saddr=(Slot-1)*WSlotSize+1
        call rffti(N2, wsave(Saddr))
c
c Get partitioning info
        call pSplit(m,msz,mlo,mhi)
c
c Initalize address mapping for compression
        naddr(1)=1
        do 145 i=2, N2/2+1
           naddr(i)=(i-1)*2
 145    continue
c
c 1. Step: Zero out dout:
        do 100 j=1, N2*M
           dout(j)=0.0D0
 100    continue
c
c 2. Step: copy din to dout
c 2.1 Copy first part
        do 110 j=1, M
           do 110 i=1, N
              dout((j-1)*N2+i)=din(i,j)
 110    continue
c 2.2 Copy wrapped-around part
        do 115 j=1, M
           do 115 i=2, N
              dout((j-1)*N2+N2-i+2)=din(i,j)
 115    continue
cc        call PrtArray(dout, N2, M, N2, 'CV1DInit-dout:')  ! <-
c
c 3. Step: fft in 1.st direction
        do 120 j=mlo+1, mhi+1
           call rfftf(N2, dout((j-1)*N2+1), wsave(saddr))
 120    continue
cc        call PrtArray(dout, N2, M, N2, 'CV1DInit-1.st fft:')  ! <-
c
c 4. Step: Merge the fft'ed data
        call pMerge(dout,m,n2)
c
c Get rid of the imaginary parts which are 0 due to symmetry
c
c Initalize address mapping for compression
        do 150 j=1, M
           do 150 i=1, N2/2+1
              dout((j-1)*(N2/2+1)+i)=dout((j-1)*N2+naddr(i))
 150    continue
        sz = (N2/2+1)*M
c
        return
      end
c CV1DInit end.
c
c
      Subroutine CV1D(d,N,M,NM,if,Mif,dout,i0,j0,NMout,scr,lscr)
C Take the data in d (declared d(1:NM, 1:M), used as d(1:N,1:M)) and
C convolute it with the function in if using FFT, with convolution in the first
C indices only. The result will be returned in dout. if will not be changed,
C so it can be used several times. if has to be prepared with CV1DInit.
C Mif is the number of fft'ed if's in if, which is prepared as if(N2,Mif) by
C CV1DInit. The size of the result in dout is always (1:N,1:mif), with
C dout being declared as (1:NMOut,1:M+j0-1)
C scr is a scratch array of at least size 4*N*M if to be used, as signaled
C by lscr=1. If lscr=0, local scratch will be allocated.
      implicit none
      integer N,M,NM,Mif,i0,j0,NMout,lscr
      real*8 d(1:NM, 1:M), if(2*N*M), dout(1:NMOut,1:Mif+j0-1)
      real*8 scr(4*N*M*lscr)
      integer sz
      real*8 dtmp1(0:2*N*M*(1-lscr)),dtmp2(0:2*N*M*(1-lscr))
c
        sz=2*N*M+1
        if(lscr.eq.1)
     &    call CV1Dpriv(d,N,M,NM,if,Mif,dout,i0,j0,NMout,scr,scr(sz))
        if(lscr.eq.0)
     &    call CV1Dpriv(d,N,M,NM,if,Mif,dout,i0,j0,NMout,dtmp1,dtmp2)
        return
      end
c
c
c
      Subroutine CV1Dpriv(d,N,M,NM,if,Mif,dout,i0,j0,NMout,dtmp1,dtmp2)
C This is the private implementation of CV1D, don't call directly.
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer N,M,NM, Mif,i0,j0,NMout
      real*8 d(1:NM, 1:M), if(2*N*Mif),dout(1:NMOut,1:Mif+j0-1)
      real*8 dtmp1(2*N*M), dtmp2(2*N*Mif), scal
      integer i,j,N2,offs,offs2,Slot,Saddr,sub,rem
      integer adr,adr2,naddr(NMax)
      integer WSlots, WSlotSize
      parameter (WSlots=2, WSlotSize=2*NMax+15)
      integer wsz(WSlots), Wused(WSlots)
      real*8 wsave(WSlotSize*WSlots)
      integer msz,mlo,mhi,mifsz,miflo,mifhi
      common /CV1Dcommon/ Wsz, wsave, Wused
c
c Div. Init stuff:
        Call CVSizes(N,1,N2,1,-1)
        Slot=1
        if( Wsz(Slot).ne.N2 ) Slot=2
        Saddr=(Slot-1)*WSlotSize+1
        if( Wsz(Slot).ne.N2 ) then
c           print *, 'Warning - Not enough Slots in CV1DPRIV'
           Slot=1
           if(Wused(Slot).eq.1) Slot=2
           Saddr=(Slot-1)*WSlotSize+1
           call rffti(N2, wsave(Saddr))
           Wsz(Slot)=N2
        endif
        Wused(1)=0
        Wused(2)=0
        Wused(Slot)=1
c
c
        do 90 i=1, N2
           naddr(i)=i/2+1
 90     continue
c
c Get partitioning info
        call pSplit(M,msz,mlo,mhi)
        call pSplit(Mif,mifsz,miflo,mifhi)
c
c 1. Step: Initialize tmp storage to zero
        do 100 i=1, N2*M
           dtmp1(i)=0.0D0
           dtmp2(i)=0.0D0
 100    continue
c
c 2. Step: copy d to dtmp1
        do 120 j=1, M
           do 120 i=1, N
              dtmp1((j-1)*N2+i)=d(i,j)
 120    continue
cc        call PrtArray(dtmp1, N2, M, N2, 'cv2d-d inserted')  ! <-
c
c 3. Step: FFT dtmp1
        do 130 j=mlo+1, mhi+1
           call rfftf(N2, dtmp1((j-1)*N2+1), wsave(Saddr))
 130    continue
cc        call PrtArray(dtmp1, N2, M, N2, 'cv2d-1st dim FFT fwd')  ! <-
c
c 4. Step: Multiply arrays
        call pMerge(dtmp1, M, N2)
        do 160 j=miflo+1, mifhi+1
           offs = mod(j-1, M)*N2
           offs2 = (j-1)*(N2/2+1)
           do 160 i=1, N2
              adr = (j-1)*N2+i
              adr2= offs2+naddr(i)
              dtmp2(adr) = dtmp1(offs+i)*if(adr2)
 160    continue
cc       call PrtArray(dtmp2, N2, Mif, N2, 'cv2d-Multiply')  ! <-
c
c 5. Step: FFT 1. Dimension back
        do 190 j=miflo+1, mifhi+1
           call rfftb(N2, dtmp2((j-1)*N2+1), wsave(Saddr))
 190    continue
cc        call PrtArray(dtmp2, N2, Mif, N2, 'cv2d-1st dim FFT back')  ! <-
c
c 6. Step: Copy relevant elements of dtmp2 to dout
        call pMerge(dtmp2, Mif, N2)
        scal=N2
        scal=1/scal
        do 200 j=1, Mif
           do 200 i=1, N
              dout(i+i0-1,j+j0-1) = dtmp2((j-1)*N2+i)*scal
 200    continue
c
c Uff. 7. Step: Go Home.
        return
      end
c CV1D End
c
c
      Subroutine CV1DGather(d,N,M,NM,if,fi1,s1,fi2,s2,scr,lscr)
C Take the data in d (declared d(1:NM, 1:M), used as d(1:N,1:M)) and
C convolute it with the function in if using FFT, with convolution in the first
C indices only. fi1 will contain the sum of the M convolutions done, which
C is a vector of N values, stored with stride s1. fi2 and s2 serve the same
C purpose, but with d mirrored in the second dimension. if will not be changed,
C so it can be used several times. if has to be prepared with CV1DInit, as
C if(N2,M).
C scr is a scratch array of at least size 2*N*M if to be used, as signaled
C by lscr=1. If lscr=0, local scratch will be allocated.
C This highly specialized operation is required to efficiently
C compute the potential contribution of the volume charges to the surfaces.
      implicit none
      integer N,M,NM,s1,s2,lscr
      real*8 d(1:NM, 1:M), if(2*N*M), fi1(N*s1), fi2(N*s2)
      real*8 scr(2*N*M*lscr)
      real*8 sz
      real*8 dtmp1(0:2*N*M*(1-lscr))
c
        sz=2*N*M+1
        if(lscr.eq.1)
     &    call CV1DGatherpriv(d,N,M,NM,if,fi1,s1,fi2,s2,scr)
        if(lscr.eq.0)
     &    call CV1DGatherpriv(d,N,M,NM,if,fi1,s1,fi2,s2,dtmp1)
        return
      end
c
c
c
      Subroutine CV1DGatherpriv(d,N,M,NM,if,fi1,s1,fi2,s2,dtmp1)
C This is the private implementation of CV1DGather, don't call directly.
      implicit none
      include 'parallel.inc'
      integer NMax
      parameter (NMax=512)
      integer N,M,NM, s1,s2
      real*8 d(1:NM, 1:M), if(2*N*M), fi1(s1*N), fi2(s1*N)
      real*8 dtmp1(2*N*M), scal
      integer i,j,N2,offs,Slot,Saddr
      integer adr,naddr(NMax)
      real*8 ddir(2*N), dflip(2*N)
      integer WSlots, WSlotSize
      parameter (WSlots=2, WSlotSize=2*NMax+15)
      integer wsz(WSlots), Wused(WSlots)
      real*8 wsave(WSlotSize*WSlots)
      integer msz,mlo,mhi
      common /CV1Dcommon/ Wsz, wsave, Wused
c
c Div. Init stuff:
        Call CVSizes(N,1,N2,1,-1)
        Slot=1
        if( Wsz(Slot).ne.N2 ) Slot=2
        Saddr=(Slot-1)*WSlotSize+1
        if( Wsz(Slot).ne.N2 ) then
c           print *, 'Warning - Not enough Slots in CV1DGatherPriv'
           Slot=1
           if(Wused(Slot).eq.1) Slot=2
           Saddr=(Slot-1)*WSlotSize+1
           call rffti(N2, wsave(Saddr))
           Wsz(Slot)=N2
        endif
        Wused(1)=0
        Wused(2)=0
        Wused(Slot)=1
c
c Initialize address Mapping
        do 90 i=1, N2
           naddr(i)=i/2+1
 90     continue
c
c Get partitioning info
        call pSplit(m,msz,mlo,mhi)
c
c 1. Step: Initialize tmp storage to zero
        do 100 i=1, N2*M
           dtmp1(i)=0.0D0
 100    continue
        do 110 i=1, N2
           ddir(i)=0
           dflip(i)=0
 110    continue
c
c 2. Step: copy d to dtmp1
        do 120 j=1, M
           do 120 i=1, N
              dtmp1((j-1)*N2+i)=d(i,j)
 120    continue
cc        call PrtArray(dtmp1, N2, M, N2, 'cv2d-d inserted')  ! <-
c
c 3. Step: FFT dtmp1
        do 130 j=mlo+1, mhi+1
           call rfftf(N2, dtmp1((j-1)*N2+1), wsave(Saddr))
 130    continue
cc        call PrtArray(dtmp1, N2, M, N2, 'cv2d-1st dim FFT fwd')  ! <-
c
c 4. Step: Multiply arrays
c NOTE: HOW can this pe parallelized efficiently? The task is to
C pointwise multiply dtmp
        call pMerge(dtmp1, M, N2)
        do 160 j=1, M
           offs = (j-1)*(N2/2+1)
           do 160 i=1, N2
              adr  = offs+naddr(i)
              ddir(i) = ddir(i) + dtmp1((j-1)*N2+i)*if(adr)
              dflip(i) = dflip(i) + dtmp1((M-j)*N2+i)*if(adr)
 160    continue
c
c 5. Step: FFT 1. Dimension back
        call rfftb(N2, ddir, wsave(Saddr))
        call rfftb(N2, dflip, wsave(Saddr))
cc        call PrtArray(dtmp2, N2, Mif, N2, 'cv2d-1st dim FFT back')  ! <-
c
c 6. Step: Copy relevant elements of dtmp2 to dout
        scal=N2
        scal=1/scal
        do 200 i=1, N
           fi1(i*s1) = ddir(i)*scal
           fi2(i*s2) = dflip(i)*scal
 200    continue
c
c Uff. 7. Step: Go Home.
        return
      end
c CV1D End
c
c
      Subroutine CVSizes(N,M,N2,M2,sz)
C Compute 'nice' Sizes for FFT, that is N2.ge.N*2 and N2=2^A*3^B*5^C;
C Same for M2. Still, the Interaction coefficients have to fit in
C an array of sz real*8. NOTE: N2 and M2 must be even!
C Input: integer N,M,sz
C Output: integer N,M
      implicit none
      integer N,M,N2,M2,sz
c
c For the moment, do nothing:
        N2=N*2
        M2=M*2
c
        return
      end
c
c
c
      Subroutine CVSizes3D(H,K,L,H2,K2,L2,sz)
C Compute 'nice' Sizes for FFT, that is N2.ge.N*2 and N2=2^A*3^B*5^C;
C Same for M2. Still, the Interaction coefficients have to fit in
C an array of sz real*8. NOTE: N2 and M2 must be even!
C Input: integer N,M,sz
C Output: integer N,M
      implicit none
      integer H,K,L,H2,K2,L2,sz
c
c For the moment, do nothing:
        H2=H*2
        K2=K*2
        L2=L*2
c
        return
      end
