C---.----1----.----2----.----3----.----4----.----5----.----6----.----7----.----
C 
C proutines.f: Routines for using MagFEM on a parallel machine.
C     These routines have been written and used such that
C     the program can still be used on a serial machine by
C     providing dummy versions of the routines.
C
C Authors: P. Lockey and K. Ramstoeck, 18. Oct. 1995
C
C==============================================================================
c
      Subroutine pStart()
      implicit none
      include 'parallel.inc'
      integer i,j,ierror
c
c Wake up the parallel nodes.      
      call MPI_init(ierror)
      call MPI_comm_rank(MPI_comm_world,myid,ierror)
      call MPI_comm_size(MPI_comm_world,nprocs,ierror)

c Initialise the partitioning information.
      do 10 i = 1, maxslots
        do 20 j = 0, maxprocs-1
          psz(i,j) = 0
          plo(i,j) = 0
          phi(i,j) = 0
 20     continue
 10   continue
      pslot = 0
      return
      end

c ------------------------------------------------------------------------------

      integer function pIndex()
      implicit none
      include 'parallel.inc'
c
c Who am I? 
c      call MPI_comm_rank(MPI_comm_world,myid,ierror)
      pIndex = myid
      end

c ------------------------------------------------------------------------------

      integer function pProcs()
      implicit none
      include 'parallel.inc'
c How many nodes do we have?
c      call MPI_comm_size(MPI_comm_world,nprocs,ierror)
      pProcs = nprocs
      end

c ------------------------------------------------------------------------------

      Subroutine pIntCast(iarray,count)
      implicit none
      include 'parallel.inc'
      integer count, ierror
      integer iarray(count)

      call MPI_bcast(iarray,count,MPI_integer,0,MPI_comm_world,
     &     ierror)

      end
c ------------------------------------------------------------------------------

      Subroutine pRealCast(rarray,count)
      implicit none
      include 'parallel.inc'
      integer count, ierror
      real*8 rarray(count)

      call MPI_bcast(rarray,count,MPI_real8,0,MPI_comm_world,
     &     ierror)

      end
c ------------------------------------------------------------------------------

      Subroutine pStrCast(sarray,count)
      include 'parallel.inc'
      integer count, ierror
      character sarray(count)

      call MPI_bcast(sarray,count,MPI_character,0,MPI_comm_world,
     &     ierror)

      end

c ------------------------------------------------------------------------------
c
      Subroutine psplit(idim,isz,ilo,ihi)
      implicit none
      include 'parallel.inc'
      integer idim,isz,ilo,ihi,sub,rem,i

c Does this splitting already exist?
      do 20 i = 1, maxslots
        if ( phi(i,nprocs-1).eq.(idim-1) ) then
          isz = psz(i,myid)
          ihi = phi(i,myid)
          ilo = plo(i,myid)
          return
        endif
 20   continue

c Need a new splitting: Which slot should it go in?
      pslot = pslot + 1
      if ( pslot .gt. maxslots ) then
        pslot = 1
      endif

c Partitioning of work over an array of dimension idim
      sub = idim / nprocs
      rem = idim - (nprocs*sub)
      do 10 i = 0, nprocs-1
         if ( i .lt. rem ) then
            psz(pslot,i) = sub + 1
            plo(pslot,i) = i*(sub+1)
         else
            psz(pslot,i) = sub
            plo(pslot,i) = rem*(sub+1) + (i-rem)*sub
         endif
         phi(pslot,i) = plo(pslot,i) + psz(pslot,i) - 1
 10   continue

c      write(*,*) '--- Proc ',myid,' Split of ',idim,' ---'
c      write(*,*) 'sz ',(psz(pslot,i),i=0,nprocs-1)
c      write(*,*) 'lo ',(plo(pslot,i),i=0,nprocs-1)
c      write(*,*) 'hi ',(phi(pslot,i),i=0,nprocs-1)

c Return partitioning data about this processor.
      isz = psz(pslot,myid)
      ihi = phi(pslot,myid)
      ilo = plo(pslot,myid)

      return
      end

c ------------------------------------------------------------------------------
C pmerge: Amalgamate results from the parallel nodes
C
c     The matrix is replicated on all nodes. Each node
c     carries out computation on a small part of the
c     replicated matrix, then all the values on the
c     sub-parts are gathered together, so that every
c     node has the values over the full matrix.
c
c     The matrices are stored in row-major order, and
c     each node is allocated a block of adjacent rows
c     to work on. The splitting up of the rows between
c     the nodes is passed in MALO & MASZ.
c
c     NOTE1: malo contains the *offset* of the first
c     row in the block for each processor (numbered
c     from 0, not 1). masz contains the number of rows
c     in each processors block.
c     NOTE2: This routine carries out the merge in a
c     very simple fashion. The merge has to be done 
c     *in place*, so MPI gather routines can not be 
c     used. There is large scope for improving this 
c     routine.
c     NOTE3: The routine carries out a merge on REAL*8
c     data. To merge COMPLEX*16 arrays, call this
c     routine with the argument na doubled.
C
C Authors: P. Lockey and K. Ramstoeck, 18. Oct. 1995
c ------------------------------------------------------------------------------
C
      Subroutine pmerge(a,ma,na)
      implicit none
      include 'parallel.inc'
      integer ma, na, row, id, rcvid, ierror, i,j,
     &     masz(0:nprocs-1), mahi(0:nprocs-1), malo(0:nprocs-1),
     &     status(MPI_status_size), request
      real*8 a(ma*na)
      logical found

c Call pSplit to ensure that one of the partitioning info
c slots has a splitting for this grid.
      call pSplit(ma,masz(0),malo(0),mahi(0))

c pSplit only returns the partitioning info for this
C processor: discard it and grab a full copy of the 
c partitioning info.
      found = .false.
      do 20 i = 1, maxslots
        if ( phi(i,nprocs-1).eq.(ma-1) ) then
          do 25 j = 0, nprocs-1
            masz(j) = psz(i,j)
            mahi(j) = phi(i,j)
            malo(j) = plo(i,j)
            found = .true.
 25       continue
        endif
 20   continue
      if ( .not. found ) then
         write(*,*) 'PMERGE: Could not find partitioning slot'
      endif

c Define an MPI type for a single row of the matrix.
      call MPI_type_contiguous(na,MPI_real8,row,ierror)
      call MPI_type_commit(row,ierror)


      do 10 id = 0, nprocs-1
         if ( myid.eq.id ) then
            do 15 rcvid = 0, nprocs-1
               if ( rcvid.ne.myid ) then
c                  write(*,*) 'Node ',myid,' sending size ',masz(myid),
c     &                 ' to ',rcvid
                  call MPI_isend(a(malo(myid)*na+1),masz(myid),row,
     &                 rcvid,TAG_pmerge,MPI_comm_world,request,ierror)
            endif
 15      continue
          else 
c             write(*,*) 'Node ',myid,' rcving size ',masz(id),
c     &            ' from ',id
            call MPI_recv(a(malo(id)*na+1),masz(id),row,id,TAG_pmerge,
     &           MPI_comm_world,status,ierror)
          endif
 10   continue

      return
      end

c ------------------------------------------------------------------------------

      Subroutine pGetCaseName(Name)
      implicit none
      character*80 Name
      integer pIndex
c
      Name=' '
      if(pindex().ne.0) return

      open(Unit=30, File='pcase.dat', status='old', Err=1000)
      read(30,*) name
      return
c
 1000 continue
      call Die('pGetCaseName', 'Couldn''t open file pcase.dat')
      return
      end
